#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Minimum distance line between two lines
"""

import volmdlr as vm
import numpy as npy

p11 = vm.Point3D(npy.random.random(3))
p12 = vm.Point3D(npy.random.random(3))
p21 = vm.Point3D(npy.random.random(3))
p22 = vm.Point3D(npy.random.random(3))

l1 = vm.Line3D(p11, p12)
l2 = vm.Line3D(p21, p22)

pmd1, pmd2 = l1.MinimumDistancePoints(l2)

u = p12 - p11 # vector of line1
v = p22 - p21 # vector of line2
w = pmd2 - pmd1

print(u.Dot(w), v.Dot(w))

m = vm.VolumeModel([('', [l1, l2, pmd1, pmd2])])
m.MPLPlot()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 11 15:16:33 2018

@author: steven
"""

import volmdlr as vm
import volmdlr.primitives3D as primitives3D
import numpy as npy
import random
p1 = vm.Point3D((0, 0, 0))
p2 = vm.Point3D((-0.150, 0, 0))
p3 = vm.Point3D((-0.150, 0.215, 0))
p4 = vm.Point3D((-0.150, 0.215, -0.058))
p5 = vm.Point3D((-0.175, 0.186, -0.042))

points = [p1, p2, p3, p4, p5]
radius = {1: 0.015, 2: 0.020, 3: 0.005}

current_point = p5.vector
#points = [p1, p2]
#radius = {1: 0.010}
for i in range(14):
    current_point += 0.300 * (npy.random.random(3) -0.5)
    points.append(vm.Point3D(current_point))
    radius[4+i] = 0.01 + 0.03 * random.random()
#print(radius)
c = vm.Circle3D(p1, 0.008, p2-p1)

rl = primitives3D.RoundedLineSegments3D(points, radius, closed=False, adapt_radius=True, name='wire')
contour = vm.Contour3D([c])

sweep = primitives3D.Sweep(contour, rl, name = 'Random pipe')

m = vm.VolumeModel([('Random Pipe', [sweep])])

m.FreeCADExport('sweep')




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 14 13:55:59 2017

@author: steven
"""

# Debug of area and inertia

import volmdlr as vm

r=0.05

p1=vm.Point2D((0,0))
p2=vm.Point2D((0,-r))
p3=vm.Point2D((r,0))
p4=vm.Point2D((0,r))

c1=vm.Arc2D(p2,p3,p4)

print(c1.SecondMomentArea(p1))

c2=vm.Circle2D(p1,r)

print(c2.SecondMomentArea(p1))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 28 14:08:23 2017

@author: steven
"""

import math

import numpy as npy
npy.seterr(divide='raise')

import volmdlr
from volmdlr.primitives import RoundedLineSegments


import matplotlib.pyplot as plt

class RoundedLineSegments3D(volmdlr.Wire3D, RoundedLineSegments):
    def __init__(self, points, radius, closed=False, adapt_radius=False, name=''):
        primitives = RoundedLineSegments.__init__(self, points, radius,
                                                  volmdlr.LineSegment3D,
                                                  volmdlr.Arc3D,
                                                  closed, adapt_radius, name='')
        volmdlr.Wire3D.__init__(self, primitives, name)

    def ArcFeatures(self, ipoint):
        radius = self.radius[ipoint]
        if self.closed:
            if ipoint == 0:
                pt1 = self.points[-1]
            else:
                pt1 = self.points[ipoint -1]
            pti = self.points[ipoint]
            if ipoint < self.npoints-1:
                pt2 = self.points[ipoint+1]
            else:
                pt2 = self.points[0]
        else:
            pt1 = self.points[ipoint - 1]
            pti = self.points[ipoint]
            pt2 = self.points[ipoint + 1]

        dist1 = (pt1 - pti).Norm()
        dist2 = (pt2 - pti).Norm()
        dist3 = (pt1 - pt2).Norm()
        alpha = math.acos(-(dist3**2-dist1**2-dist2**2)/(2*dist1*dist2))/2.
        dist = radius/math.tan(alpha)

        u1 = (pt1 - pti) / dist1
        u2 = (pt2 - pti) / dist2

        p3 = pti + u1*dist
        p4 = pti + u2*dist

        n = u1.Cross(u2)
        n /= n.Norm()
        v1 = u1.Cross(n)
        v2 = u2.Cross(n)

        l1 = volmdlr.Line3D(p3, p3+v1)
        l2 = volmdlr.Line3D(p4, p4+v2)
        c, _ = l1.MinimumDistancePoints(l2)

        u3 = u1 + u2# mean of v1 and v2
        u3 /= u3.Norm()

        interior = c - u3 * radius
        return p3, interior, p4, dist, alpha


    def Rotation(self, center, angle, copy=True):
        if copy:
            return RoundedLineSegments3D([p.Rotation(center, angle, copy=True)\
                                          for p in self.points],
                                         self.radius, self.closed, self.name)
        else:
            self.__init__([p.Rotation(center, angle, copy=True)\
                           for p in self.points],
                          self.radius, self.closed, self.name)

    def Translation(self, offset, copy=True):
        if copy:
            return RoundedLineSegments3D([p.Translation(offset, copy=True)\
                                          for p in self.points],
                                         self.radius, self.closed, self.name)
        else:
            self.__init__([p.Translation(offset, copy=True)\
                           for p in self.points],
                          self.radius, self.closed, self.name)

class Sphere(volmdlr.Primitive3D):
    def __init__(self,center, radius, name=''):
        volmdlr.Primitive3D.__init__(self, name)
        self.center = center
        self.radius = radius
        self.position = center

    def Volume(self):
        return 4/3*math.pi*self.radius**3

    def FreeCADExport(self, ip, ndigits=3):
        name = 'primitive'+str(ip)
        r = 1000*self.radius
        x, y, z = npy.round(1000*self.center.vector, ndigits)
        return '{} = Part.makeSphere({}, fc.Vector({}, {}, {}))\n'.format(name,r,x,y,z)

class Block(volmdlr.Shell3D):
    """
    Creates a block
    :param frame: a frame 3D. The origin of the frame is the center of the block,
     the 3 vectors are defining the edges. The frame has not to be orthogonal
    """
    def __init__(self, frame, name='', color=None):
        self.frame = frame
        self.size = (self.frame.u.Norm(), self.frame.v.Norm(), self.frame.w.Norm())
        
        faces = self.shell_faces()
        volmdlr.Shell3D.__init__(self, faces, name, color)

    def Vertices(self):
        return [self.frame.origin - 0.5*self.frame.u - 0.5*self.frame.v - 0.5*self.frame.w,
                self.frame.origin - 0.5*self.frame.u + 0.5*self.frame.v - 0.5*self.frame.w,
                self.frame.origin + 0.5*self.frame.u + 0.5*self.frame.v - 0.5*self.frame.w,
                self.frame.origin + 0.5*self.frame.u - 0.5*self.frame.v - 0.5*self.frame.w,
                self.frame.origin - 0.5*self.frame.u - 0.5*self.frame.v + 0.5*self.frame.w,
                self.frame.origin - 0.5*self.frame.u + 0.5*self.frame.v + 0.5*self.frame.w,
                self.frame.origin + 0.5*self.frame.u + 0.5*self.frame.v + 0.5*self.frame.w,
                self.frame.origin + 0.5*self.frame.u - 0.5*self.frame.v + 0.5*self.frame.w]

    def Edges(self):
        p1, p2, p3, p4, p5, p6, p7, p8 = self.Vertices()
        return [volmdlr.LineSegment3D(p1, p2),
                volmdlr.LineSegment3D(p2, p3),
                volmdlr.LineSegment3D(p3, p4),
                volmdlr.LineSegment3D(p4, p1),
                volmdlr.LineSegment3D(p5, p6),
                volmdlr.LineSegment3D(p6, p7),
                volmdlr.LineSegment3D(p7, p8),
                volmdlr.LineSegment3D(p8, p5),
                volmdlr.LineSegment3D(p1, p5),
                volmdlr.LineSegment3D(p2, p6),
                volmdlr.LineSegment3D(p3, p7),
                volmdlr.LineSegment3D(p4, p8)]
        
    def face_contours(self):
        e1, e2, e3, e4, e5, e6, e7, e8, e9, e10, e11, e12 = self.Edges()
        return [volmdlr.Contour3D([e1, e2, e3, e4]),
                volmdlr.Contour3D([e5, e6, e7, e8]),
                volmdlr.Contour3D([e1, e9, e5, e10]),
                volmdlr.Contour3D([e2, e10, e6, e11]),
                volmdlr.Contour3D([e3, e11, e7, e12]),
                volmdlr.Contour3D([e4, e12, e8, e9])]
        
    def shell_faces(self):
        c1, c2, c3, c4, c5, c6 = self.face_contours()
        return [volmdlr.Face3D([c1]),
                volmdlr.Face3D([c2]),
                volmdlr.Face3D([c3]),
                volmdlr.Face3D([c4]),
                volmdlr.Face3D([c5]),
                volmdlr.Face3D([c6])]

    def plot_data(self, x3D, y3D, marker=None, color='black', stroke_width=1,
                  dash=False, opacity=1, arrow=False):
        lines = []
        for edge3D in self.Edges():
            lines.append(edge3D.plot_data(x3D, y3D, marker, color, stroke_width,
                         dash, opacity, arrow))

        return lines

    def MPLPlot2D(self, x3D, y3D, ax=None):
        if ax is None:
            fig, ax = plt.subplots()
            ax.set_aspect('equal')
        else:
            fig = None

        for edge3D in self.Edges():
#            edge2D = edge3D.PlaneProjection2D()
            edge3D.MPLPlot2D(x3D, y3D, ax)

        return fig, ax

class Cylinder(volmdlr.Primitive3D):
    def __init__(self, position, axis, radius, length, name=''):
        volmdlr.Primitive3D.__init__(self, name)
        self.position = position
        axis.Normalize()
        self.axis = axis
        self.radius = radius
        self.length = length
        self.bounding_box = self._bounding_box()
        
    def _bounding_box(self):
        pointA = self.position - self.length/2 * self.axis
        pointB = self.position + self.length/2 * self.axis
        
        dx2 = (pointA[0]-pointB[0])**2
        dy2 = (pointA[1]-pointB[1])**2
        dz2 = (pointA[2]-pointB[2])**2
        
        kx = ((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5
        ky = ((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5
        kz = ((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5
        
        if pointA[0] > pointB[0]:
            pointA, pointB = pointB, pointA
        xmin = pointA[0] - kx * self.radius
        xmax = pointB[0] + kx * self.radius
        
        if pointA[1] > pointB[1]:
            pointA, pointB = pointB, pointA
        ymin = pointA[1] - ky * self.radius
        ymax = pointB[1] + ky * self.radius
        
        if pointA[2] > pointB[2]:
            pointA, pointB = pointB, pointA
        zmin = pointA[2] - kz * self.radius
        zmax = pointB[2] + kz * self.radius
        
        return volmdlr.BoundingBox(xmin, xmax, ymin, ymax, zmin, zmax)

    def Volume(self):
        return self.length * math.pi * self.radius**2

    def FreeCADExport(self, ip):
        if self.radius > 0:
            name = 'primitive'+str(ip)
            e = str(1000*self.length)
            r = str(1000*self.radius)
            position = 1000*(self.position - self.axis*self.length/2.)
            x, y, z = position
            x = str(x)
            y = str(y)
            z = str(z)

            ax, ay, az = self.axis
            ax = str(ax)
            ay = str(ay)
            az = str(az)
            return name+'=Part.makeCylinder('+r+','+e+',fc.Vector('+x+','+y+','+z+'),fc.Vector('+ax+','+ay+','+az+'),360)\n'
        else:
            return ''

    def Babylon(self):
        new_axis = volmdlr.Vector3D((self.axis[0], self.axis[1], self.axis[2]))
        normal_vector1 = new_axis.RandomUnitNormalVector()
        normal_vector2 = new_axis.Cross(normal_vector1)
        x, y, z = self.position
        s='var cylinder = BABYLON.Mesh.CreateCylinder("{}", {}, {}, {}, 30, 1, scene,false, BABYLON.Mesh.DEFAULTSIDE);'.format(self.name,self.length,2*self.radius,2*self.radius)
        s+='cylinder.position = new BABYLON.Vector3({},{},{});\n;'.format(x,y,z)
        s += 'var axis1 = new BABYLON.Vector3({},{},{});\n'.format(new_axis[0], new_axis[1], new_axis[2])
        s += 'var axis2 = new BABYLON.Vector3({},{},{});\n'.format(normal_vector1[0], normal_vector1[1], normal_vector1[2])
        s += 'var axis3 = new BABYLON.Vector3({},{},{});\n'.format(normal_vector2[0], normal_vector2[1], normal_vector2[2])
        s += 'cylinder.rotation = BABYLON.Vector3.RotationFromAxis(axis3, axis1, axis2);\n'
        return s

class HollowCylinder(volmdlr.Primitive3D):
    def __init__(self, position, axis, inner_radius, outer_radius, width, name=''):
        volmdlr.Primitive3D.__init__(self, name)
        self.position = position
        axis.Normalize()
        self.axis = axis
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.width = width

    def Volume(self):
        return self.width * math.pi* (self.outer_radius**2 - self.inner_radius**2)


    def FreeCADExport(self, ip):
        if self.outer_radius > 0.:
            name = 'primitive'+str(ip)
            re = round(1000*self.outer_radius, 6)
            ri = round(1000*self.inner_radius, 6)
            x, y, z = round((1000*(self.position - self.axis*self.width/2)), 6)
            ax, ay, az = npy.round(self.axis.vector, 6)

            s='C2 = Part.makeCircle({}, fc.Vector({}, {}, {}),fc.Vector({}, {}, {}))\n'.format(re, x, y, z, ax, ay, az)
            s+='W2 = Part.Wire(C2.Edges)\n'
            s+='F2 = Part.Face(W2)\n'

            if self.inner_radius!=0.:
                s+='C1 = Part.makeCircle({}, fc.Vector({}, {}, {}),fc.Vector({}, {}, {}))\n'.format(ri, x, y, z, ax, ay, az)
                s+='W1 = Part.Wire(C1.Edges)\n'
                s+='F1 = Part.Face(W1)\n'
                s+='F2 = F2.cut(F1)\n'

            vx, vy, vz = round(self.axis*self.width*1000, 6)

            s += '{} = F2.extrude(fc.Vector({}, {}, {}))\n'.format(name, vx, vy, vz)
            return s

        else:
            return ''

    def Babylon(self):
        ya,xa,za=self.axis# to counter y definition in babylon
        theta=math.acos(za/self.width)
        phi=math.atan(ya/xa)
        x,z,y=self.position
        s='var cylinder = BABYLON.Mesh.CreateCylinder("{}", {}, {}, {}, 30, 1, scene,false, BABYLON.Mesh.DEFAULTSIDE);'.format(self.name,self.width,2*self.outer_radius,2*self.outer_radius)
        s+='cylinder.position = new BABYLON.Vector3({},{},{});\n;'.format(x,y,z)
        s+='cylinder.rotation.x={}\n;'.format(-theta*math.sin(phi))
        s+='cylinder.rotation.y={}\n;'.format(theta*math.cos(phi))
        s+='cylinder.rotation.z={}\n;'.format(phi)
        return s
    

class Cone(volmdlr.Primitive3D):
    def __init__(self, position, axis, radius, length, name=''):
        volmdlr.Primitive3D.__init__(self, name)
        self.position = position
        axis.Normalize()
        self.axis = axis
        self.radius = radius
        self.length = length
        self.bounding_box = self._bounding_box()
        
    def _bounding_box(self):
        """
        A is the point at the basis
        B is the top
        """
        pointA = self.position - self.length/2 * self.axis
        pointB = self.position + self.length/2 * self.axis
        
        dx2 = (pointA[0]-pointB[0])**2
        dy2 = (pointA[1]-pointB[1])**2
        dz2 = (pointA[2]-pointB[2])**2
        
        kx = ((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5
        ky = ((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5
        kz = ((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5
        
        x_bound = (pointA[0] - kx * self.radius, pointA[0] + kx * self.radius, pointB[0])
        xmin = min(x_bound)
        xmax = max(x_bound)
        
        y_bound = (pointA[1] - ky * self.radius, pointA[1] + ky * self.radius, pointB[1])
        ymin = min(y_bound)
        ymax = max(y_bound)
        
        z_bound = (pointA[2] - kz * self.radius, pointA[2] + kz * self.radius, pointB[2])
        zmin = min(z_bound)
        zmax = max(z_bound)
        
        return volmdlr.BoundingBox(xmin, xmax, ymin, ymax, zmin, zmax)

    def Volume(self):
        return self.length * math.pi * self.radius**2 / 3

    def Babylon(self):
        new_axis = volmdlr.Vector3D((self.axis[0], self.axis[1], self.axis[2]))
        normal_vector1 = new_axis.RandomUnitNormalVector()
        normal_vector2 = new_axis.Cross(normal_vector1)
        x, y, z = self.position
        s = 'var cone = BABYLON.MeshBuilder.CreateCylinder("cone", {{diameterTop:0, diameterBottom:{}, height: {}, tessellation: 100}}, scene);\n'.format(2*self.radius, self.length)
        s += 'cone.position = new BABYLON.Vector3({},{},{});\n;'.format(x,y,z)
        s += 'var axis1 = new BABYLON.Vector3({},{},{});\n'.format(new_axis[0], new_axis[1], new_axis[2])
        s += 'var axis2 = new BABYLON.Vector3({},{},{});\n'.format(normal_vector1[0], normal_vector1[1], normal_vector1[2])
        s += 'var axis3 = new BABYLON.Vector3({},{},{});\n'.format(normal_vector2[0], normal_vector2[1], normal_vector2[2])
        s += 'cone.rotation = BABYLON.Vector3.RotationFromAxis(axis3, axis1, axis2);\n'
        return s


class ExtrudedProfile(volmdlr.Primitive3D):
    """

    """
    def __init__(self, plane_origin, x, y, outer_contour2d, inner_contours2d,
                 extrusion_vector, name=''):
        volmdlr.Primitive3D.__init__(self, name)
        self.outer_contour2d = outer_contour2d
        self.outer_contour3d = outer_contour2d.To3D(plane_origin, x, y)
        self.inner_contours2d = inner_contours2d
        self.extrusion_vector = extrusion_vector
        self.inner_contours3d = []
        self.x = x
        self.y = y

        bool_areas = []
        for contour in inner_contours2d:
            self.inner_contours3d.append(contour.To3D(plane_origin, x, y))
            if contour.Area() > outer_contour2d.Area():
                bool_areas.append(True)
            else:
                bool_areas.append(False)
        if any(bool_areas):
            raise ValueError('At least one inner contour is not contained in outer_contour.')

    def MPLPlot(self, ax):
        for contour in self.contours3D:
            for primitive in contour:
                primitive.MPLPlot(ax)

    def FreeCADExport(self, ip):
        name='primitive'+str(ip)
        s = 'Wo = []\n'
        s += 'Eo = []\n'
        for ip, primitive in enumerate(self.outer_contour3d.basis_primitives):
            s += primitive.FreeCADExport('L{}'.format(ip))
            s += 'Eo.append(Part.Edge(L{}))\n'.format(ip)
        s += 'Wo.append(Part.Wire(Eo[:]))\n'
        s += 'Fo = Part.Face(Wo)\n'

        s += 'Fi = []\n'
        s += 'W = []\n'
        for ic,contour in enumerate(self.inner_contours3d):
            s+='E = []\n'
            for ip, primitive in enumerate(contour.basis_primitives):
                s += primitive.FreeCADExport('L{}_{}'.format(ic, ip))
                s += 'E.append(Part.Edge(L{}_{}))\n'.format(ic, ip)
            s += 'Wi = Part.Wire(E[:])\n'
            s += 'Fi.append(Part.Face(Wi))\n'

        if len(self.inner_contours3d) != 0:
            s += 'Fo = Fo.cut(Fi)\n'
        e1, e2, e3 = round(1000*self.extrusion_vector, 6)

        s+='{} = Fo.extrude(fc.Vector({}, {}, {}))\n'.format(name,e1,e2,e3)
        return s

    def Area(self):
        areas = [c.Area() for c in self.contours2D]
        sic=list(npy.argsort(areas))[::-1]# sorted indices of contours
        area=areas[sic[0]]

        for i in sic[1:]:
            area-=self.contours2D[i].Area()
        return area

    def Volume(self):
#        e=extrusion_vector/norm(extrusion_vector)
        z = npy.cross(self.x.vector,self.y.vector)
        z.Normalize()
        coeff = npy.dot(self.extrusion_vector, z)

        return self.Area()*coeff


class RevolvedProfile(volmdlr.Primitive3D):
    """

    """
    def __init__(self, plane_origin, x, y, contours2D, axis_point,
                 axis,angle=2*math.pi, name=''):
        volmdlr.Primitive3D.__init__(self, name)
        self.contours2D = contours2D
        self.axis_point = axis_point
        self.axis = axis
        self.angle = angle
        self.plane_origin = plane_origin
        self.x = x
        self.y = y

        self.contours3D = []
        for contour in contours2D:
            self.contours3D.append(contour.To3D(plane_origin, x, y))

    def MPLPlot(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots()
        for contour in self.contours3D:
#            for primitive in contour:
            contour.MPLPlot(ax)

    def FreeCADExport(self, ip, ndigits=3):
        name = 'primitive'+str(ip)
        s = 'W=[]\n'
        for ic, contour in enumerate(self.contours3D):
            s += 'L=[]\n'
            for ibp, basis_primitive in enumerate(contour.basis_primitives):
                s += basis_primitive.FreeCADExport('L{}_{}'.format(ic, ibp), 8)
                s += 'L.append(L{}_{})\n'.format(ic,ibp)
            s += 'S = Part.Shape(L)\n'
            s += 'W.append(Part.Wire(S.Edges))\n'
        s += 'F=Part.Face(W)\n'
        a1, a2, a3 = self.axis.vector
        ap1, ap2, ap3 = self.axis_point.vector
        ap1 = round(ap1*1000, ndigits)
        ap2 = round(ap2*1000, ndigits)
        ap3 = round(ap3*1000, ndigits)
        angle = self.angle/math.pi*180
        s += '{} = F.revolve(fc.Vector({},{},{}), fc.Vector({},{},{}),{})\n'.format(name, ap1,ap2,ap3,a1,a2,a3,angle)

#            myObject.Shape = Sweep
        return s

    def Volume(self):
        areas=[c.Area() for c in self.contours2D]
        # Maximum area is main surface, others cut into it
        sic = list(npy.argsort(areas))[::-1]# sorted indices of contours
        p1=self.axis_point.PlaneProjection3D(self.plane_origin,self.x,self.y)
        if self.axis_point.PointDistance(p1)!=0:
            raise NotImplementedError
        p1_2D=p1.To2D(self.axis_point,self.x,self.y)
        p2_3D=self.axis_point+volmdlr.Point3D(self.axis.vector)
        p2=p2_3D.PlaneProjection3D(self.plane_origin,self.x,self.y)
        if p2_3D.PointDistance(p2)!=0:
            raise NotImplementedError
        p2_2D=p2_3D.To2D(self.plane_origin,self.x,self.y)
        axis_2D=volmdlr.Line2D(p1_2D,p2_2D)
        com = self.contours2D[sic[0]].CenterOfMass()
        rg = axis_2D.PointDistance(com)
        volume=areas[sic[0]]*rg

        for i in sic[1:]:
            com=self.contours2D[i].CenterOfMass()
            rg=axis_2D.PointDistance(com)
            volume-=areas[i]*rg

        return self.angle*volume



class HelicalExtrudedProfile(volmdlr.Primitive3D):
    """

    """
    def __init__(self, plane_origin, x, y, axis_point,axis, pitch,
                 outer_contour2d, inner_contours2d=None, name=''):
        volmdlr.Primitive3D.__init__(self, name)
        if inner_contours2d is not None:
            self.inner_contours2d = inner_contours2d
        else:
            self.inner_contours2d = []
        self.outer_contour2d = outer_contour2d
        self.axis_point=axis_point
        self.axis=axis
        self.pitch=pitch

        self.inner_contours3d=[c.To3D(plane_origin,x,y) for c in inner_contours2d]
        self.outer_contour3d=outer_contour2d.To3D(plane_origin,x,y)


    def FreeCADExport(self,ip,ndigits=3):
        name='primitive{}'.format(ip)
        s="E = []\n"
        for icontour, contour in enumerate(self.outer_contour3d.basis_primitives):
            s += contour.FreeCADExport('L_{}'.format(icontour))
            s += 'E.append(Part.Edge(L_{}))\n'.format(icontour)
        s += 'W = Part.Wire(E[:])\n'

#        a1,a2,a3=self.axis
        ap1, ap2, ap3 = self.axis_point
        ap1 = round(ap1*1000, ndigits)
        ap2 = round(ap2*1000, ndigits)
        ap3 = round(ap3*1000, ndigits)

        width = self.axis.Norm()*1000
        direction = bool(self.pitch < 0)
        pitch = round(abs(self.pitch)*1000, ndigits)
        s += "helix = Part.makeHelix({}, {}, 50., 0, {})\n".format(pitch, width, direction)
        s += "helix.translate(fc.Vector({},{},{}))\n".format(ap1, ap2, ap3)

        s += '{} = helix.makePipeShell([W],True,True)\n'.format(name)
        for ic, contour in enumerate(self.inner_contours3d):
            s += "Ei=[]\n"
            s += "helix2 = Part.makeHelix({}, {}, 50., 0, {})\n".format(pitch,1.01*width+pitch,direction)
            s += "helix2.translate(fc.Vector({},{},{}))\n".format(ap1,ap2,ap3-pitch)
            for ip, primitive in enumerate(contour.basis_primitives):
                s += primitive.FreeCADExport('L_{}_{}'.format(ic,ip))
                s += 'Ei.append(Part.Edge(L_{}_{}))\n'.format(ic,ip)
            s+= 'Wi = Part.Wire(Ei[:])\n'
            s+= "{} = {}.cut(helix2.makePipeShell([Wi],True,True))\n".format(name,name)

        return s


class Sweep(volmdlr.Primitive3D):
    """
    Sweep a 3D contour along a Wire3D
    """
    def __init__(self, contour3d, wire3d, name=''):
        volmdlr.Primitive3D.__init__(self,name)
        self.contour3d = contour3d
        self.wire3d = wire3d

    def FreeCADExport(self, ip, ndigits=3):
        name = 'primitive{}'.format(ip)
        s = "E = []\n"
        for icontour, contour in enumerate(self.contour3d.basis_primitives):
            s += contour.FreeCADExport('L_{}'.format(icontour))
            s += 'E.append(Part.Edge(L_{}))\n'.format(icontour)
        s += 'contour = Part.Wire(E[:])\n'

        s += "E=[]\n"
        for iwire, wire in enumerate(self.wire3d.basis_primitives):
            s += wire.FreeCADExport('L_{}'.format(iwire))
            s += 'E.append(Part.Edge(L_{}))\n'.format(iwire)
        s += 'wire = Part.Wire(E[:])\n'

        s += '{} = wire.makePipeShell([contour],True, True)\n'.format(name)


        return s

class Cut(volmdlr.Primitive3D):
    """
    Cut primitive 1 by primitive 2
    """
    def __init__(self,primitive,cut_primitives,name=''):
        volmdlr.Primitive3D.__init__(self,name)
        self.primitive=primitive
        self.cut_primitives = cut_primitives


    def FreeCADExport(self,ip):
        name = 'primitive{}'.format(ip)

        s = self.primitive.FreeCADExport('{}'.format(ip))
        for icp, cut_primitive in enumerate(self.cut_primitives):
            s += cut_primitive.FreeCADExport('{}_{}'.format(ip, icp))
            s += "{} = {}.cut({}_{})\n".format(name, name, name, icp)

        return s

class Fuse(volmdlr.Primitive3D):
    """
    Fuse primitives
    """
    def __init__(self, primitives, name=''):
        volmdlr.Primitive3D.__init__(self, name)
        self.primitives = primitives


    def FreeCADExport(self,ip):
        name = 'primitive{}'.format(ip)


        s = self.primitives[0].FreeCADExport(ip)
        for primitive in self.primitives[1:]:
            s += primitive.FreeCADExport('{}_0'.format(ip))
            s += "{} = {}.fuse({}_0)\n".format(name,name,name)

        return s


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 28 14:07:37 2017

@author: steven
"""

import math
import numpy as npy
npy.seterr(divide='raise')
import volmdlr as vm
#from itertools import permutations
import jsonschema
import json

import matplotlib.pyplot as plt
from matplotlib.patches import Arc, FancyArrow
from mpl_toolkits.mplot3d import Axes3D
import pkg_resources
import tempfile
import webbrowser

from jinja2 import Environment, PackageLoader, select_autoescape

_jsonschema = {
    "definitions": {},
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "plot_data Base Schema",
    'type' : 'array',
    'editable' : True,
    'description' : 'plot data',
    "items" : {
          "anyOf": [
               {"type" : "object",
                "required": ['type', 'data', 'color', 'marker', 'size', 'dash'],
                "properties": {'type' : {"const": 'line'},
                               'data' : {'type' : 'array',
                                         "minItems": 4,
                                         "maxItems": 4,
                                         'examples' : [[0, 1, 1, 2]],
                                         "description" : "Point coord",
                                         'items' : {'type' : 'number',
                                                    'step' : "any",
                                                    'unit' : 'm'
                                                    }
                                         },
                                'color': {"enum": ['black', 'red', 'blue']},
                                'marker': {"enum": [None, 'o']},
                                'dash': {'type': 'boolean'},
                                'size': {"type": 'number',
                                         "examples" : [0.5],
                                         "step" : 'any',
                                         "minimum" : 0,
                                         "description" : "Point size"},
                                'opacity': {"type": 'number',
                                            "examples" : [1],
                                            "step" : 'any',
                                            "minimum" : 0,
                                            "description" : "Point opacity"},
                               }
                       },
               {"type" : "object",
                "required": ['type', 'data', 'color', 'marker', 'size'],
                "properties": {'type' : {"const": 'point'},
                               'data' : {'type' : 'array',
                                         "minItems": 2,
                                         "maxItems": 2,
                                         'examples' : [[0, 1]],
                                         "description" : "Point coord",
                                         'items' : {'type' : 'number',
                                                    'step' : "any",
                                                    'unit' : 'm'
                                                    }
                                         },
                                'color': {"enum": ['black', 'red', 'blue']},
                                'marker': {"enum": ['o']},
                                'size': {"type": 'number',
                                         "examples" : [0.5],
                                         "step" : 'any',
                                         "minimum" : 0,
                                         "description" : "Point size"},
                                'opacity': {"type": 'number',
                                            "examples" : [1],
                                            "step" : 'any',
                                            "minimum" : 0,
                                            "description" : "Point opacity"},
                               }
                       },
               {"type" : "object",
                "properties": {'type' : {"const": 'contour'}}
                       },
               {"type" : "object",
                "properties": {'type' : {"const": 'wire'}}
                       },
                    ]
                    },
         }
     
         
color = {'black': 'k', 'blue': 'b', 'red': 'r', 'green': 'g'}

def validate(plot_datas):
    return jsonschema.validate(instance=plot_datas, schema=_jsonschema)

def plot_d3(plot_datas):
    env = Environment(loader=PackageLoader('volmdlr', 'templates'),
                          autoescape=select_autoescape(['html', 'xml']))
    template = env.get_template('plot_data.html')
    
    volmdlr_path = pkg_resources.resource_filename(pkg_resources.Requirement('volmdlr'),
                                              'volmdlr/templates')
            
    s = template.render(
                        volmdlr_path=volmdlr_path,
                        D3Data=json.dumps(plot_datas))

    temp_file = tempfile.mkstemp(suffix='.html')[1]
    
    with open(temp_file, 'wb') as file:
        file.write(s.encode('utf-8'))

    webbrowser.open('file://' + temp_file)
    print('file://' + temp_file)

def plot(plot_datas, ax=None):
    if ax is None:
        fig, ax = plt.subplots()
        ax.set_aspect('equal')
    else:
        fig = None
        
    for plot_data in plot_datas:
        if plot_data['type'] == 'line':
            style = ''
            if plot_data['dash']:
                style += '--'
            else:
                style += '-'
            style += color[plot_data['color']]
            p1, p2 = plot_data['data'][0: 2], plot_data['data'][2:] 
            if plot_data['arrow']:
                ax.plot([p1[0], p2[0]], [p1[1], p2[1]], style)
                length = ((p1[0] - p2[0])**2 + (p1[1] - p2[1])**2)**0.5
                if width is None:
                    width = length / 1000.
                    head_length = length/20.
                    head_width = head_length/2.
                else:
                    head_width = 2*width
                    head_length = head_width
                ax.arrow(p1[0], p1[1], (p2[0] - p1[0])/length*(length - head_length),
                         (p2[1] - p1[1])/length*(length - head_length),
                         head_width = head_width, fc = 'b', linewidth = 0,
                         head_length = head_length, width = width, alpha = 0.3)
            else:
                ax.plot([p1[0], p2[0]], [p1[1], p2[1]], style, linewidth=plot_data['size'])
                
        elif plot_data['type'] == 'point':
            p1 = plot_data['data']
            style = ''
            style += color[plot_data['color']]
            style += plot_data['marker']
            ax.plot(p1[0], p1[1], style, linewidth=plot_data['size'])
            
        elif plot_data['type'] == 'contour':
            plot(plot_data['plot_data'], ax)
            
        elif plot_data['type'] == 'arc':
            pc = vm.Point2D(( plot_data['cx'],  plot_data['cy']))
            ax.add_patch(Arc(pc, 2*plot_data['r'], 2*plot_data['r'], angle=0,
                        theta1=plot_data['angle1']*0.5/math.pi*360,
                        theta2=plot_data['angle2']*0.5/math.pi*360,
                        color=color[plot_data['color']], linewidth=plot_data['size']))
            
        elif plot_data['type'] == 'circle':
            pc = vm.Point2D(( plot_data['cx'],  plot_data['cy']))
            ax.add_patch(Arc(pc, 2*plot_data['r'], 2*plot_data['r'], angle=0,
                        theta1=0,
                        theta2=360,
                        color=color[plot_data['color']], linewidth=plot_data['size']))
    return fig, ax

    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  3 09:56:29 2017

@author: steven
"""

import numpy as npy
import volmdlr as vm
import volmdlr.primitives3D as primitives3D
import volmdlr.primitives2D as primitives2D

p1=vm.Point2D((0, 0))
p2=vm.Point2D((0.1, 0.))
p3=vm.Point2D((0.1, 0.2))
p4=vm.Point2D((0.05, 0.1))
p5=vm.Point2D((0.,0.21))
p6=vm.Point2D((0.05, 0.05))

p7 = vm.Point2D((0.06, 0.05))
p8 = vm.Point2D((0.04, 0.07))

radius = {0: 0.01, 2: 0.01, 3: 0.015}

outer_profile = primitives2D.RoundedLineSegments2D([p1, p2, p3, p4, p5], radius, closed = True)
#hole = vm.Circle2D(p6, 0.01)
#inner_profile = primitives2D.RoundedLineSegments2D([p6, p7, p8], {0: }, closed = True)
l1 = vm.LineSegment2D(p6, p7)
l2 = vm.LineSegment2D(p7, p8)
l3 = vm.LineSegment2D(p8, p6)
c2 = vm.Contour2D([l1,l2,l3])

c1 = vm.Contour2D([outer_profile])
#c2 = vm.Contour2D([inner_profile])

po=vm.Point3D((0, 0, 0))
xp=vm.Vector3D((1, 0, 0))
yp=vm.Vector3D((0, 1, 0))
f, a = c1.MPLPlot()
c2.MPLPlot(a)

profile=primitives3D.ExtrudedProfile(po, xp, yp, c1, [c2], vm.Vector3D((0,0, 0.2)), name = 'extrusion')

model=vm.VolumeModel([('profile', [profile])])

#profile.MPLPlot((0,0,0),(1,0,0),(0,1,0))

#model.MPLPlot()

model.FreeCADExport('extrusion')


# -*- coding: utf-8 -*-
"""
Setup install script for volmdlr

"""

from setuptools import setup
#from distutils.core import setup
from Cython.Build import cythonize

from os.path import dirname, isdir, join
import re
from subprocess import CalledProcessError, check_output

tag_re = re.compile(r'\btag: %s([0-9][^,]*)\b')
version_re = re.compile('^Version: (.+)$', re.M)

def readme():
    with open('README.rst') as f:
        return f.read()
    
def get_version():
    # Return the version if it has been injected into the file by git-archive
    version = tag_re.search('$Format:%D$')
    if version:
        return version.group(1)

    d = dirname(__file__)
    
    if isdir(join(d, '.git')):
        cmd = 'git describe --tags'
        try:
            version = check_output(cmd.split()).decode().strip()[:]
        except CalledProcessError:
            raise RuntimeError('Unable to get version number from git tags')
        if version[0]=='v':
            version = version[1:]
        # PEP 440 compatibility
        if '-' in version:
            future_version = version.split('-')[0].split('.')
            future_version[-1] = str(int(future_version[-1])+1)
            future_version = '.'.join(future_version)
            number_commits = version.split('-')[1]
            version = '{}.dev{}'.format(future_version, number_commits)
            return version

    else:
        # Extract the version from the PKG-INFO file.
        with open(join(d, 'PKG-INFO')) as f:
            version = version_re.search(f.read()).group(1)
            
#    # Writing to file
#    with open('powertransmission/version.py', 'w+') as vf:
#        vf.write("# -*- coding: utf-8 -*-\nversion = '{}'".format(version))
                 
    return version


setup(name='volmdlr',
      version = get_version(),
#      setup_requires=['setuptools_scm'],
      description=' A volume modeler computation-oriented. Include rendering bindings. ',
      long_description=readme(),
      keywords='volume, modeler',
      url='https://github.com/masfaraud/volmdlr',
      author='Steven Masfaraud',
      author_email='steven@masfaraud.fr',
      license='Creative Commons Attribution-Share Alike license',
      packages=['volmdlr'],#,'volmdlr.primitives2D','volmdlr.primitives3D','volmdlr.geometry'],
      package_dir={},
      install_requires=['numpy', 'matplotlib', 'Cython', 'scipy', 'jinja2',
                        'geomdl', 'jsonschema'],
      classifiers=['Topic :: Scientific/Engineering','Development Status :: 3 - Alpha'],
      ext_modules = cythonize("volmdlr/vmcy.pyx"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""

import math
import volmdlr
from volmdlr.primitives import RoundedLineSegments


class RoundedLineSegments2D(volmdlr.Wire2D, RoundedLineSegments):
    def __init__(self, points, radius, closed=False, adapt_radius=False, name=''):
        primitives = RoundedLineSegments.__init__(self, points, radius,
                                                  volmdlr.LineSegment2D, volmdlr.Arc2D,
                                                  closed, adapt_radius, name='')

        volmdlr.Wire2D.__init__(self, primitives, name)


    def ArcFeatures(self, ipoint):
        radius = self.radius[ipoint]
        if self.closed:
            if ipoint == 0:
                pt1 = self.points[-1]
            else:
                pt1 = self.points[ipoint -1]
            pti = self.points[ipoint]
            if ipoint < self.npoints-1:
                pt2 = self.points[ipoint+1]
            else:
                pt2 = self.points[0]
        else:
            pt1 = self.points[ipoint - 1]
            pti = self.points[ipoint]
            pt2 = self.points[ipoint + 1]

        # TODO: change to PointDistance
        dist1 = (pt1-pti).Norm()
        dist2 = (pt2-pti).Norm()
        dist3 = (pt1-pt2).Norm()
        alpha = math.acos(-(dist3**2 - dist1**2 - dist2**2) / (2*dist1*dist2)) / 2.
        dist = radius / math.tan(alpha)

        u1 = (pt1-pti) / dist1
        u2 = (pt2-pti) / dist2

        p3 = pti + u1 * dist
        p4 = pti + u2 * dist

        w = (u1+u2)
        if w != volmdlr.Vector2D((0, 0)):
            w.Normalize()

        v1 = u1.NormalVector()
        if v1.Dot(w) < 0:
            v1 = -v1

        pc = p3 + v1 * radius
        pm = pc - radius*w

        return p3, pm, p4, dist, alpha


    def Rotation(self, center, angle, copy=True):
        if copy:
            return RoundedLineSegments2D([p.Rotation(center, angle, copy=True)\
                                          for p in self.points],
                                          self.radius, self.closed,
                                          adapt_radius =self.adapt_radius,
                                          name = self.name)
        else:
            self.__init__([p.Rotation(center, angle, copy=True) for p in self.points],
                           self.radius, self.closed,
                           adapt_radius=self.adapt_radius, name = self.name)

    def Translation(self, offset, copy=True):
        if copy:
            return RoundedLineSegments2D([p.Translation(offset, copy=True) for p in self.points],
                                          self.radius, self.closed, adapt_radius=self.adapt_radius, name = self.name)
        else:
            self.__init__([p.Translation(offset,copy=True) for p in self.points],
                           self.radius, self.closed, adapt_radius =self.adapt_radius, name = self.name)

    def Offset(self, offset):
        nb = len(self.points)
        vectors = []
        for i in range(nb-1):
            v1 = volmdlr.Vector2D((self.points[i+1] - self.points[i]))
            v2 = volmdlr.Vector2D((self.points[i] - self.points[i+1]))
            v1.Normalize()
            v2.Normalize()
            vectors.append(v1)
            vectors.append(v2)

        if self.closed:
            v1 = volmdlr.Vector2D((self.points[0] - self.points[-1]))
            v2 = volmdlr.Vector2D((self.points[-1] - self.points[0]))
            v1.Normalize()
            v2.Normalize()
            vectors.append(v1)
            vectors.append(v2)


        offset_vectors = []
        new_radii = {}
        offset_points = []

        for i in range((not self.closed),nb-(not self.closed)):

            check = False
            ni = vectors[2*i-1] + vectors[2*i]
            if ni == volmdlr.Vector2D((0,0)):
                ni = vectors[2*i]
                ni = ni.NormalVector()
                offset_vectors.append(ni)
            else:
                ni.Normalize()
                if ni.Dot(vectors[2*i-1].NormalVector()) > 0:
                    ni = - ni
                    check = True
                offset_vectors.append(ni)

            if i in self.radius:
                if (check and offset > 0) or (not check and offset < 0):
                    new_radius = self.radius[i] + abs(offset)
                else:
                    new_radius = self.radius[i] - abs(offset)
                if new_radius > 0:
                    new_radii[i] = new_radius
                else:
                    if self.adapt_radius:
                        new_radii[i] = 1e-6

            normal_vector1 = - vectors[2*i-1].NormalVector()
            normal_vector2 =   vectors[ 2*i ].NormalVector()
            normal_vector1.Normalize()
            normal_vector2.Normalize()
            alpha = math.acos(normal_vector1.Dot(normal_vector2))

            offset_point = self.points[i] + offset/math.cos(alpha/2)*offset_vectors[i-(not self.closed)]
            offset_points.append(offset_point)


        if not self.closed:
            n1 = vectors[0].NormalVector(unit=True)
            offset_vectors.insert(0, n1)
            offset_points.insert(0, self.points[0] + offset*offset_vectors[0])

            n_last = vectors[-1].NormalVector(unit=True)
            n_last = - n_last
            offset_vectors.append(n_last)
            offset_points.append(self.points[-1] + offset*offset_vectors[-1])

        return RoundedLineSegments2D(offset_points, new_radii, self.closed,
                                         adapt_radius=self.adapt_radius)
        
        
    def OffsetSingleLine(self, line_index, offset):
        """
        line_index = 0 being the 1st line 
        """
        new_linesegment2D_points = []
        dont_add_last_point = False

        for i, point in enumerate(self.points[:-1]+(self.closed)*[self.points[-1]]):

            if i == line_index:
                # Not closed RLS2D and the offset line is the last one
                if i == len(self.points)-2:
                    dir_vec_1 = volmdlr.Vector2D(point-self.points[i-1])
                    dir_vec_1.Normalize()
                    dir_vec_2 = dir_vec_1
                    dont_add_last_point = True
                # The offset line is the first one
                elif i == 0:
                    dir_vec_2 = volmdlr.Vector2D(self.points[i+1]-self.points[i+2])
                    dir_vec_2.Normalize()
                    if not self.closed:
                        dir_vec_1 = dir_vec_2
                    else:
                        dir_vec_1 = volmdlr.Vector2D(point-self.points[i-1])
                        dir_vec_1.Normalize()
                # Closed RLS2D and the offset line is the last one
                elif i == len(self.points)-1:
                    dir_vec_1 = volmdlr.Vector2D(point-self.points[i-1])
                    dir_vec_1.Normalize()
                    dir_vec_2 = volmdlr.Vector2D(self.points[0]-self.points[1])
                    dir_vec_2.Normalize()
                    dont_add_last_point = True
                else:
                    dir_vec_1 = volmdlr.Vector2D(point-self.points[i-1])
                    dir_vec_1.Normalize()
                    dir_vec_2 = volmdlr.Vector2D(self.points[i+1]-self.points[i+2])
                    dir_vec_2.Normalize()
                
                
                if self.closed and line_index == len(self.points)-1:
                    normal_vector = volmdlr.Vector2D(self.points[0]-point).NormalVector(unit=True)
                else:
                    normal_vector = volmdlr.Vector2D(self.points[i+1]-point).NormalVector(unit=True)
                    
                alpha1 = math.acos(dir_vec_1.Dot(normal_vector))
                alpha2 = math.acos(dir_vec_2.Dot(normal_vector))
                
                # If 3 segments are aligned and the middle one have to be offset
                if math.isclose(math.cos(alpha1), 0, abs_tol=1e-9) or math.isclose(math.cos(alpha2), 0, abs_tol=1e-9):
                    print('ca sort direct')
                    print('direction vector', dir_vec_1, dir_vec_2)
                    print('normal vector', normal_vector)
                    print('alpha', alpha1*180/math.pi, alpha2*180/math.pi)
                    print(point, self.points[i+1])
                    return self
#                    distance_dir1 = offset
#                    distance_dir2 = offset
                    
                distance_dir1 = offset / math.cos(alpha1)
                distance_dir2 = offset / math.cos(alpha2)

                new_point1 = point + distance_dir1 * dir_vec_1
                if self.closed and line_index == len(self.points)-1:
                    new_point2 = self.points[0] + distance_dir2 * dir_vec_2
                else:
                    new_point2 = self.points[i+1] + distance_dir2 * dir_vec_2
                
                new_linesegment2D_points.append(new_point1)
                new_linesegment2D_points.append(new_point2)
                
            elif i == line_index+1:
                pass
            
            elif line_index == len(self.points)-1 and i == 0:
                pass
            else:
                new_linesegment2D_points.append(point)
            
            
        if not dont_add_last_point and not self.closed:
            new_linesegment2D_points.append(self.points[-1])
        
        rls2D = RoundedLineSegments2D(new_linesegment2D_points, self.radius,
                                                 self.closed, adapt_radius=self.adapt_radius)
        
        return rls2D
        
    
    def OffsetLines(self, line_indexes, offset):
        """
        line_indexes is a list of consecutive line indexes
        These line should all be aligned
        line_indexes = 0 being the 1st line 
        
        if self.close last line_index can be len(self.points)-1
        if not, last line_index can be len(self.points)-2
        """  
        new_linesegment2D_points = []
        
# =============================================================================
# COMPUTES THE DIRECTIVE VECTORS BETWEEN WHICH THE OFFSET WILL BE DRAWN 
# =============================================================================
        dir_vec_1 = None
        dir_vec_2 = None
            
        if line_indexes[0] == 0 and not self.closed:
            pass
        else: 
            dir_vec_1 = volmdlr.Vector2D((self.points[line_indexes[0]] - self.points[line_indexes[0]-1]))
        
        if line_indexes[-1] == len(self.points)-(2-self.closed):
            if not self.closed:
                pass
            else:
                dir_vec_2 = volmdlr.Vector2D((self.points[0] - self.points[1]))
        elif self.closed and line_indexes[-1] == len(self.points)-2:
            dir_vec_2 = volmdlr.Vector2D((self.points[line_indexes[-1]+1] - self.points[0]))
        else:
            dir_vec_2 = volmdlr.Vector2D((self.points[line_indexes[-1]+1] - self.points[line_indexes[-1]+2]))

        if dir_vec_1 is None:
            dir_vec_1 = dir_vec_2
        if dir_vec_2 is None:
            dir_vec_2 = dir_vec_1
                
        dir_vec_1.Normalize()
        dir_vec_2.Normalize()
            
    
# =============================================================================
# COMPUTES THE ANGLE BETWEEN THE NORMAL VECTOR OF THE SURFACE TO OFFSET AND 
# THE DIRETIVE VECTOR IN ORDER TO SET THE NEW POINT AT THE RIGHT DISTANCE
# =============================================================================
        normal_vectors = []
        for index in line_indexes:
            if index == len(self.points)-1:
                normal_vectors.append(volmdlr.Vector2D(self.points[0]-self.points[index]).NormalVector(unit=True))
            else:
                normal_vectors.append(volmdlr.Vector2D(self.points[index+1]-self.points[index]).NormalVector(unit=True))

        dot1 = dir_vec_1.Dot(normal_vectors[0])
        dot2 = dir_vec_2.Dot(normal_vectors[-1])
    
        if math.isclose(dot1, 0, abs_tol=1e-9):
            # call function considering the line before, because the latter and 
            # the first offset segment are parallel
            return self.OffsetLines([line_indexes[0]-1]+line_indexes, offset)
        if math.isclose(dot2, 0, abs_tol=1e-9):
            # call function considering the line after, because the latter and 
            # the last offset segment are parallel
            return self.OffsetLines(line_indexes+[line_indexes[-1]+1], offset)
        
        distance_dir1 = offset / dot1
        distance_dir2 = offset / dot2
        
        if len(line_indexes) > 1:
            intersection = volmdlr.Point2D.LinesIntersection(volmdlr.Line2D(self.points[line_indexes[0]], self.points[line_indexes[0]]+dir_vec_1), volmdlr.Line2D(self.points[line_indexes[-1]+1], self.points[line_indexes[-1]+1]+dir_vec_2))
            vec1 = intersection.PointDistance(self.points[line_indexes[0]]) * dir_vec_1
            vec2 = intersection.PointDistance(self.points[line_indexes[-1]+1]) * dir_vec_2
        
# =============================================================================
# COMPUTES THE NEW POINTS AFTER THE OFFSET
# =============================================================================
        new_points = {}
        
        new_points[line_indexes[0]] = self.points[line_indexes[0]] + distance_dir1 * dir_vec_1
        
        for nb, index in enumerate(line_indexes[1:]):
            coeff_vec_2 = volmdlr.Point2D.PointDistance(self.points[line_indexes[0]], self.points[index]) / volmdlr.Point2D.PointDistance(self.points[line_indexes[0]], self.points[line_indexes[-1]+1])
            coeff_vec_1 = 1 - coeff_vec_2
            if dir_vec_1.Dot(normal_vectors[nb+1]) < 0:
                coeff_vec_1 = - coeff_vec_1
            if dir_vec_2.Dot(normal_vectors[nb+1]) < 0:
                coeff_vec_2 = - coeff_vec_2
            index_dir_vector = coeff_vec_1 * vec1 + coeff_vec_2 * vec2
            index_dot = index_dir_vector.Dot(normal_vectors[nb+1])
            index_distance_dir = offset / index_dot
            new_points[index] = self.points[index] + index_distance_dir * index_dir_vector
            
        if self.closed and line_indexes[-1] == len(self.points)-1:
            new_points[0] = self.points[0] + distance_dir2 * dir_vec_2
        else:
            new_points[line_indexes[-1]+1] = self.points[line_indexes[-1]+1] + distance_dir2 * dir_vec_2
        
# =============================================================================
# CREATE THE NEW POINTS' LIST 
# ============================================================================= 
        for i in range(len(self.points)):
            if i in new_points.keys():
                new_linesegment2D_points.append(new_points[i])
            else:
                new_linesegment2D_points.append(self.points[i])
                
        
        rls2D = RoundedLineSegments2D(new_linesegment2D_points, self.radius,
                                                 self.closed, adapt_radius=self.adapt_radius)
        

        return rls2D


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 28 15:33:01 2017

@author: steven
"""
import numpy as npy
import volmdlr as vm
import volmdlr.primitives3D as primitives3D

cylinder1 = primitives3D.Cylinder(vm.Vector3D((0.,0.,0.)), vm.Vector3D((1.,0.,0.)),
                                0.03, 0.02, 'cylinder1')
cylinder2 = primitives3D.HollowCylinder(vm.Vector3D((0,0.1,0.)), vm.Vector3D((1.,0.,0.)),
                                      0.02, 0.06, 0.03,'cylinder2')
#profile=primitives3D.ExtrudedProfile((0,0,0),(1,0,0),(0,1,0),[(0,0),(0.1,0.),(0.15,0.4),(0.,0.3)],{0:0.05,2:0.01},(0,0,0.2))

model=vm.VolumeModel([('Cylinder 1', [cylinder1]), ('Cylinder 2', [cylinder2])])

#profile.MPLPlot((0,0,0),(1,0,0),(0,1,0))

model.FreeCADExport('cylinders')

#print(model.BabylonScript())
#model.BabylonShow()